/*
 * Load or remove a tiny module
 *
 * Author: Fabrice Bellard
 * Copyright (C) 2002-2003 Netgem S.A.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <elf.h>
#include <fcntl.h>
#include <linux/linkage.h>
#include <linux/module.h>
#include <errno.h>
#include <ctype.h>

#include <asm/unistd.h>

#include "tinsmod_load.h"

_syscall1(int, delete_module, const char *, name)

void error(const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    fprintf(stderr, "tinsmod: ");
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, "\n");
    va_end(ap);
}

void help(void)
{
    printf("tinsmod (c) 2000-2002 Netgem\n"
	   "usage: tinsmod [-h] [-r] module [module_params...]\n"
	   "\n" 
           "-h : help\n" 
           "-r : remove module\n"
           "-n : no load, just show\n");
}



int main(int argc, char **argv)
{
    int ret, c, do_rmmod = 0, do_noload = 0;
    char *filename;

    while ((c = getopt(argc, argv, "hrnv")) != -1) {
	switch (c) {
	case 'h':
	    help();
	    exit(1);
	case 'r':
	    do_rmmod = 1;
	    break;
        case 'n':
            do_noload = 1;
            break;
        case 'v':
            /* ignored for compatibility with insmod */
            break;
	}
    }

    if (optind >= argc) {
	help();
	exit(1);
    }

    filename = argv[optind];

    if (!do_rmmod) {
        return tiny_insmod(filename, argc - (optind + 1),
                           argv + (optind + 1), do_noload);
    } else {
	ret = delete_module(filename);
	if (ret < 0) {
	    perror(filename);
	}
    }

    return 0;
}
