#if !defined (__MINIX_FS_H__)
#define __MINIX_FS_H__

/* JMA: Adapted from linux/minix_fs.h to avoid including linux headers */
#include <stdint.h>

/*
 * Thanks to Kees J Bot for sending me the definitions of the new
 * minix filesystem (aka V2) with bigger inodes and 32-bit block
 * pointers.
 */

#define MINIX_ROOT_INO 1

/* Not the same as the bogus LINK_MAX in <linux/limits.h>. Oh well. */
#define MINIX_LINK_MAX	250
#define MINIX2_LINK_MAX	65530

#define MINIX_I_MAP_SLOTS	8
#define MINIX_Z_MAP_SLOTS	64
#define MINIX_SUPER_MAGIC	0x137F		/* original minix fs */
#define MINIX_SUPER_MAGIC2	0x138F		/* minix fs, 30 char names */
#define MINIX_SUPER_MAGIC2_COMPR 0x139F  /* minix V1 fs, 30 char names, compr */
#define MINIX2_SUPER_MAGIC	0x2468		/* minix V2 fs */
#define MINIX2_SUPER_MAGIC2	0x2478		/* minix V2 fs, 30 char names */
#define MINIX2_SUPER_MAGIC2_COMPR 0x2488 /* minix V2 fs, 30 char names, compr */ 
#define MINIX_VALID_FS		0x0001		/* Clean fs. */
#define MINIX_ERROR_FS		0x0002		/* fs has errors. */

#define MINIX_INODES_PER_BLOCK ((BLOCK_SIZE)/(sizeof (struct minix_inode)))
#define MINIX2_INODES_PER_BLOCK ((BLOCK_SIZE)/(sizeof (struct minix2_inode)))

#define MINIX_V1		0x0001		/* original minix fs */
#define MINIX_V2		0x0002		/* minix V2 fs */

#define INODE_VERSION(inode)	inode->i_sb->u.minix_sb.s_version

/* compression flags : only meaningful in compressed filesystem */
#define MINIX_UID_MASK         0x3FFF
#define MINIX_FLAGS_MASK       0xC000

#define MINIX_COMPRBLK_NONE    0x0000
#define MINIX_COMPRBLK_4KB     0x4000
#define MINIX_COMPRBLK_8KB     0x8000 /* not implemented */
#define MINIX_COMPRBLK_16KB    0xC000 /* not implemented */

/* magic number for the start of a compressed cluster */
#define MINIX_COMPR_MAGIC      0xc6c2

/* the files are compressed by clusters of 4KB */
#define MINIX_CLUSTER_BITS 12
#define MINIX_CLUSTER_SIZE (1 << MINIX_CLUSTER_BITS)

/* ioctls for compression flags */
#include <linux/ioctl.h>
#define	MINIX_IOC_GETFLAGS		_IOR('f', 1, long)
#define	MINIX_IOC_SETFLAGS		_IOW('f', 2, long)

/*
 * This is the original minix inode layout on disk.
 * Note the 8-bit gid and atime and ctime.
 */
struct minix_inode {
    uint16_t i_mode;
    uint16_t i_uid;
    uint32_t i_size;
    uint32_t i_time;
    uint8_t  i_gid;
    uint8_t  i_nlinks;
    uint16_t i_zone[9];
};

/*
 * The new minix inode has all the time entries, as well as
 * long block numbers and a third indirect block (7+1+1+1
 * instead of 7+1+1). Also, some previously 8-bit values are
 * now 16-bit. The inode is now 64 bytes instead of 32.
 */
struct minix2_inode {
    uint16_t i_mode;
    uint16_t i_nlinks;
    uint16_t i_uid;
    uint16_t i_gid;
    uint32_t i_size;
    uint32_t i_atime;
    uint32_t i_mtime;
    uint32_t i_ctime;
    uint32_t i_zone[10];
};

/*
 * minix super-block data on disk
 */
struct minix_super_block {
    uint16_t s_ninodes;
    uint16_t s_nzones;
    uint16_t s_imap_blocks;
    uint16_t s_zmap_blocks;
    uint16_t s_firstdatazone;
    uint16_t s_log_zone_size;
    uint32_t s_max_size;
    uint16_t s_magic;
    uint16_t s_state;
    uint32_t s_zones;
};

struct minix_dir_entry {
    uint16_t inode;
    char name[0];
};

#endif /* !defined (__MINIX_FS_H__) */
