/* 
 * Release under GPL License by Jocelyn Mayer <l_indien@magic.fr>
 * Copyright Feb 2002
 */
/* The goal of this utility is too compare,
 * as fast as possible,
 * two file, with one which can be a subpart of the other
 */
/* 
 * Usage:
 * fastcmp [ -o offset ] file1 [ -o offset ] file2
 * Returns 0 if file1 is included in file2 
 *           OR file2 is included in file1
 *        -1 if there's an error
 *         1 if the files doesn't match.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

/* Is this useful ? */
#define MAXSIZE ( 16 * 1024 * 1024 )

void usage (void)
{
    fprintf(stderr, "fastcmp [ -o offset ] file1 [ -o offset ] file2\n");
//    fprintf(stderr, "file1 OR file 2 can be - or stdin\n");
}

int main (int argc, char **argv)
{
    struct stat f_stat;
    int fd1 = -1, fd2 = -1;
    off_t off1 = 0, off2 = 0;
    ssize_t size1 = 0, size2 = 0, compare;
    int numarg = 1;
    char *file1 = NULL, *file2 = NULL;
    char *ptr1 = NULL, *ptr2 = NULL, *pos1, *pos2;
    int ret = -1;

    if (argc < 3) {
        fprintf(stderr, "Too few arguments\n");
        usage();
        goto end;
    }

    if (argv[numarg][0] == '-') {
        if (argv[numarg][1] == 'o') {
            off1 = atoi(argv[numarg + 1]);
            numarg += 2;
        } else {
#if 0
            if (argv[numarg][1] == '\0') {
                fd1 = 0;
                numarg++;
            } else 
#endif
            {
                fprintf(stderr, "Bad argument: %s\n", argv[numarg]);
                usage();
                goto end;
            }
        }
    }
    file1 = argv[numarg]; 
    numarg++;

    if (argv[numarg][0] == '-') {
        if (argv[numarg][1] == 'o') {
            off2 = atoi(argv[numarg + 1]);
            numarg += 2;
        } else {
#if 0
            if (argv[numarg][1] == '\0') {
                if (fd1 != 0) {
                    fd2 = 0;
                    numarg++;
                } else {
                    fprintf(stderr, "Both file1 and file2 can't be stdin\n");
                    usage();
                }
            } else 
#endif
            {
                fprintf(stderr, "Bad argument: %s\n", argv[numarg]);
                usage();
                goto end;
            }
        }
    }
    file2 = argv[numarg];

    /* Now, we got all usefull infos, let's go on */
    if (stat(file1, &f_stat) < 0) {
        fprintf(stderr, "Can't get infos on file: %s\n", file1);
        goto end;
    }
    size1 = f_stat.st_size;
    if (stat(file2, &f_stat) < 0) {
        fprintf(stderr, "Can't get infos on file: %s\n", file2);
        goto end;
    }
    size2 = f_stat.st_size;
    if (off1 > size1 || off2 > size2) {
        fprintf(stderr, "Offsets should be < file size !\n");
        goto end;
    }

    if (fd1 < 0 && (fd1 = open(file1, O_RDONLY | O_EXCL | O_SYNC)) < 0) {
        fprintf(stderr, "Can't open file: %s\n", file1);
        goto end;
    }
    if (fd2 < 0 && (fd2 = open(file2, O_RDONLY | O_EXCL | O_SYNC)) < 0) {
        fprintf(stderr, "Can't open file: %s\n", file2);
        close(fd1);
        goto end;
    }
    if ((ptr1 = mmap(NULL, size1, PROT_READ, MAP_PRIVATE, fd1, 0)) == NULL) {
        fprintf(stderr, "Can't map file: %s\n", file1);
        goto end;
    }
    if ((ptr2 = mmap(NULL, size2, PROT_READ, MAP_PRIVATE, fd2, 0)) == NULL) {
        fprintf(stderr, "Can't map file: %s\n", file2);
        goto end;
    }
    
    pos1 = ptr1 + off1;
    size1 -= off1;
    pos2 = ptr2 + off2;
    size2 -= off2;
    compare = (size1 < size2) ? size1 : size2;
#if 0
    fprintf(stderr, "Got to compare %d bytes between %s (%ld b - o %ld) "
            "and %s (%ld b - o %ld)\n", compare, file1, size1 + off1, off1,
            file2, size2 + off2, off2);
#endif

    ret = 0;
    while (compare > 0 && ret == 0) {
        if (compare > MAXSIZE) {
            ret = memcmp(pos1, pos2, MAXSIZE);
            compare -= MAXSIZE;
        } else {
            ret = memcmp(pos1, pos2, compare);
            compare = 0;
        }
    }

 end:
    if (ptr2 != NULL)
        munmap(ptr2, size2);
    if (ptr1 != NULL)
        munmap(ptr1, size1);

    if (fd2 > 2)
        close(fd2);
    if (fd1 > 2)
        close(fd1);

    return ret;
}
