/*
 *   sdpl.h: all common sdpl definitions
 *
 *   Copyright (C) 2001 - 2007, Jocelyn MAYER <l_indien at magic dot fr>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Simple Disk Partition Layer
 *
 * SDPL common definitions
 *
 */

#ifndef __SDPL_H__
#define __SDPL_H__

/* Main infos */
#define SDPL_NAME  "Simple Disk Partition Layer"

#define SDPL_VERSION 0x00B
#define SDPL_VERSION_STR "1.0 Beta3"

/* 60 - 63 are reserved for experimentation */
#define SDPL_MAJOR 60
#define SDPL_DEVNAME "sl"

/* Useful definitions... */
#ifndef FALSE
#define FALSE 0
#endif /* FALSE */
#ifndef TRUE
#define TRUE 1
#endif /* TRUE */

#ifndef UNUSED
#define UNUSED __attribute__((unused))
#endif

#if (LINUX_VERSION_CODE == 0x20409)
#undef min
#undef max
#endif /* (LINUX_VERSION_CODE == 0x20409) */
#ifndef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif /* min */
#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif /* max */

#ifndef __KERNEL__
typedef __u8  u8;
typedef __u16 u16;
typedef __u32 u32;
typedef __u64 u64;
typedef __s8  s8;
typedef __s16 s16;
typedef __s32 s32;
typedef __s64 s64;
#endif

#if !defined(CONFIG_BLK_SDPL_DEBUG)
#define CONFIG_BLK_SDPL_DEBUG 0
#endif
#if !defined(CONFIG_BLK_SDPL_EMBEDDED)
#define CONFIG_BLK_SDPL_EMBEDDED 0
#endif
#if !defined(CONFIG_BLK_SDPL_EDIT)
#define CONFIG_BLK_SDPL_EDIT 0
#endif
#if !defined(CONFIG_SDPL_USE_MBR_COPY)
#define CONFIG_SDPL_USE_MBR_COPY 0
#endif

/* For debugging (seems that it's needed for User-mode Linux) */
/* Force it for User-Mode Linux */
#ifdef CONFIG_BLK_SDPL_USE_UML
#undef NO_PERM_CHECK
#define NO_PERM_CHECK 1
#endif

/* May be parameters in non-embedded case ? */
#if CONFIG_BLK_SDPL_EMBEDDED
#define MAX_MEDIUM 1
#define MEDIUM_SHIFT 3
#define MAX_MINOR 4
#else /* CONFIG_BLK_SDPL_EMBEDDED */
#define MAX_MEDIUM 4
#define MEDIUM_SHIFT 5
#define MAX_MINOR (1 << MEDIUM_SHIFT)
#endif /* CONFIG_BLK_SDPL_EMBEDDED */

#define MASK_MINOR ((1 << MEDIUM_SHIFT) - 1)

#include <linux/version.h>
#if (LINUX_VERSION_CODE < 0x2040A)
#ifdef CONFIG_BLK_SDPL_64BITS
#undef CONFIG_BLK_SDPL_64BITS
#endif /* HAS_64BITS_SUPPORT */
#define CONFIG_BLK_SDPL_64BITS 0
#endif /* (LINUX_VERSION_CODE < 0x2040A) */

/* Internal type definition */
#if CONFIG_BLK_SDPL_EMBEDDED

typedef u16 std_type;
typedef s16 std_stype;
#define STD_ATTRIB "%04u"
#ifdef CONFIG_BLK_SDPL_64BITS
#undef CONFIG_BLK_SDPL_64BITS
#endif /* HAS_64BITS_SUPPORT */
#define CONFIG_BLK_SDPL_64BITS 0

#ifdef CONFIG_BLK_SDPL_STATS
#undef CONFIG_BLK_SDPL_STATS
#endif /*  CONFIG_BLK_SDPL_STATS */
#define CONFIG_BLK_SDPL_STATS 0

#else /* CONFIG_BLK_SDPL_EMBEDDED */

#if CONFIG_BLK_SDPL_64BITS
typedef u64 std_type;
typedef s64 std_stype;
#define STD_ATTRIB "%04llu"
#else /* HAS_64BITS_SUPPORT */
typedef u32 std_type;
typedef s32 std_stype;
#define STD_ATTRIB "%04u"
#endif /* HAS_64BITS_SUPPORT */

#ifdef CONFIG_BLK_SDPL_STATS
#undef CONFIG_BLK_SDPL_STATS
#endif /*  CONFIG_BLK_SDPL_STATS */
#define CONFIG_BLK_SDPL_STATS 1

#endif /* CONFIG_BLK_SDPL_EMBEDDED */

/* Structure used to store infos on the bloc device */
/* Partition entry definitions */
struct SDPL_pentry16 {
    u16  magic;
    u16  flags;
    char   res0[4];
    /* 0x08 */
    u16  size;
    u16  next_part;
    u16  map_offset;
    char   res1[2];
    /* 0x10 */
} __attribute__((packed));
struct SDPL_pentry32 {
    u16 magic;
    u16  flags;
    char   res0[4];
    /* 0x08 */
    u32  size;
    char   res1[4];
    /* 0x10 */
    u32  next_part;
    u32  map_offset;
    char   res2[8];
    /* 0x20 */
} __attribute__((packed));
struct SDPL_pentry64 {
    u16  magic;
    u16  flags;
    char   res0[4];
    /* 0x08 */
    u64  size;
    /* 0x10 */
    u64  next_part;
    u64  map_offset;
    /* 0x20 */
} __attribute__((packed));

/* SDPL Info bloc definition */
/* Embedded version (16 bits) */
struct SDPL_infos16 {
    u16 bloc_size;
    u16 SDPL_size;
    u16 nparts;
    u16 table_offset;
    u16 bitmap_offset;
    u16 embedded;
    u16 data_offset;
} __attribute__((packed));
/* Common version (32 bits) */
struct SDPL_infos32 {
    u32 bloc_size;
    u32 SDPL_size;
    u32 nparts;
    u32 table_offset;
    u32 bitmap_offset;
    u32 embedded;
    u32 data_offset;
} __attribute__((packed));
/* Server version (64 bits) */
struct SDPL_infos64 {
    u64 bloc_size;
    u64 SDPL_size;
    u64 nparts;
    u64 table_offset;
    u64 bitmap_offset;
    u64 embedded;
    u64 data_offset;
} __attribute__((packed));

/* Boot Record description */
typedef struct SDPL_br_t SDPL_br_t;
struct SDPL_br_t {
    /* 0x000 */
    char  jump[4];
    u32 CRC;
    /* 0x008 */
    char  entry_buf[64];
    /* 0x048 */
    char  bufmap[320];
    /* 0x188 */
    char  res0[8];
    /* 0x190 */
    char  bitmap[32];
    /* 0x1B0 */
    char  res1[14];
    /* 0x1BE */
    u16 info_size;
    /* 0x1C0 */
    char  SDPL_infos[56];
    char  res2[4];
    u16 magic;
    u16 reserved; /* In case we want MS-DOS compatibility ? */
} __attribute__((packed));

/* Structures used to represent the layer in memory */
/* MBR infos structures */
union SDPL_infos {
    struct SDPL_infos16 infos16;
#if !CONFIG_BLK_SDPL_EMBEDDED
    struct SDPL_infos32 infos32;
#if CONFIG_BLK_SDPL_64BITS
    struct SDPL_infos64 infos64;
#endif /* CONFIG_BLK_SDPL_64BITS */
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
};

/* Partition entries */
union part_entry {
    struct SDPL_pentry16 pentry16;
#if !CONFIG_BLK_SDPL_EMBEDDED
    struct SDPL_pentry32 pentry32;
#if CONFIG_BLK_SDPL_64BITS
    struct SDPL_pentry64 pentry64;
#endif /* CONFIG_BLK_SDPL_64BITS */
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
};

/* Global Internal datas structure */
struct SDPL_map {
    /* MBR memory mapped pointer */
    SDPL_br_t *mbr;
    /* Main layer's infos */
    union SDPL_infos *infos;
    /* Real bloc device size & layer size */
    std_type device_size;
    std_type st_size;
    u32 CRC;
    u16 magic;
    u16 info_size;
#if !CONFIG_BLK_SDPL_EMBEDDED
    u64 res_size;
    std_type nparts;
    u16 embedded; /* Flag for embedded maps */
    u16 pad; /* Just to stay aligned... */
#else
    u32 res_size;
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
    std_type bloc_size;
    std_type bloc_dec;
    std_type SDPL_size;
    std_type pentries_offset;
    std_type bitmap_offset;
    std_type data_offset;
    /* Partitions entries */
    union part_entry **pentries;
    u16 *pmagic;
    u16 *pflags;
    std_type *psize;
    std_type *poffset;
    std_type *pnext;
    u16 *bflags;
    std_type **pbloc;
};

#if CONFIG_BLK_SDPL_EMBEDDED
#define NPARTS(SDPL_map) 4
#else /* CONFIG_BLK_SDPL_EMBEDDED */
#define NPARTS(SDPL_map) (SDPL_map->nparts)
#define EMBEDDED(SDPL_map) (SDPL_map->embedded)
#endif /* CONFIG_BLK_SDPL_EMBEDDED */

/* Bloc internal flags */
#define EMPTY_BLOC 0xFFFF
#define EMPTY_FILL 0xFF

/* Partition flags definition */
/* We let 8 bits as private for the OS */
#define SDPL_PRIVATE 0xFF00
#define SDPL_PUBLIC  0x00FF
#define SDPL_PUBLIC_BITS 8
#define SDPL_RO      0x00
#define SDPL_RW      0x01
#define SDPL_BOOT    0x80
typedef struct {
    std_type partnum;
    u8     flags;
} priv_flags_t;

/* SDPL ioctls */
enum {
    SDPL_ATTACH = 0x80, SDPL_DETACH, SDPL_SET_ACCESS, SDPL_GET_ACCESS,
    SDPL_ADD_BLOCS, SDPL_FREE_BLOC, SDPL_SET_BOOTABLE, SDPL_GET_BOOTABLE,
    SDPL_SET_BLOC_ACCESS, SDPL_GET_BLOC_ACCESS, SDPL_SET_PRIVATE_FLAGS,
    SDPL_GET_PRIVATE_FLAGS, SDPL_GET_BLOCSIZE, SDPL_GET_RESSIZE,
    SDPL_GET_NUMPARTS,
#if !CONFIG_BLK_SDPL_EMBEDDED
    SDPL_ADD_PARTITION, SDPL_REMOVE_PARTITION,
#endif
    SDPL_GET_VERSION = 0xA0,
};

/* Kernel space definitions */
#ifdef __KERNEL__

#ifndef CONFIG_BLK_SDPL_EDIT
#define CONFIG_BLK_SDPL_EDIT 1
#endif

/* It seems that this declaration is missing in linux headers */
kdev_t name_to_kdev_t(char *line);

/*
 * This is from fs/partition/check.h
 * and is needed to read a sector
 */
typedef struct { struct page *v; } Sector;

typedef struct layer_t layer_t;
typedef struct SDPL_device_t SDPL_device_t;
struct SDPL_device_t {
    /* Layer back reference */
    layer_t *layer;
    /* Keep procfs name */
#ifdef CONFIG_PROC_FS
    char proc_name[16];
#endif
    std_type map_entry;
    /* For statistics */
#if CONFIG_BLK_SDPL_STATS
    ulong reads, writes;
    ulong opens;
#endif /* CONFIG_BLK_SDPL_STATS */
    /* Access mode */
    int access;
    /* Should make intelligent use of this !!! */
    spinlock_t lock;
};

struct layer_t {
    SDPL_device_t *SDPL_dev;
    /* Real device */
    kdev_t device;
    /* Shared map */
    struct SDPL_map *SDPL_map;
    spinlock_t medium_lock;
    std_type minors;
    std_type max;
};

/* Internal function definitions */
#define sdpl_malloc(a) ({ kmalloc(a, GFP_KERNEL); })
#define sdpl_free(a) ({ kfree(a); })

#ifdef CONFIG_PROC_FS
#define REGISTER_PROCFS() ({ register_procfs(); })
#define ADD_PROC_ENTRY(medium, minor) \
do { sdpl_add_entry(medium, minor); } while (0)
#define REMOVE_PROC_ENTRY(medium, minor) \
do { remove_proc_entry(SDPL_DEV(medium, minor).proc_name, 0); } while (0)
#define CLEANUP_PROCFS()  do { cleanup_procfs(); } while (0)
#else /* CONFIG_PROC_FS */
#define REGISTER_PROCFS() (0)
#define ADD_PROC_ENTRY(medium, minor) do { } while (0)
#define REMOVE_PROC_ENTRY(medium, minor) do { } while (0)
#define CLEANUP_PROCFS()  do { } while (0)
#endif /* CONFIG_PROC_FS */

#ifdef CONFIG_DEVFS_FS
#define REGISTER_DEVFS()  ({ register_devfs(); })
#define CLEANUP_DEVFS()   do { cleanup_devfs(); } while (0)
#else /* CONFIG_DEVFS_FS */
#define REGISTER_DEVFS()  (0)
#define CLEANUP_DEVFS()  do { } while (0)
#endif /* CONFIG_DEVFS_FS */

#if CONFIG_BLK_SDPL_EMBEDDED

#define SDPL_LAYER(medium) (&SDPL_layer)
#define INIT_DEVS(medium) do { } while(0)
#define FOR_EACH_MEDIUM medium = 0;
#define IS_DEV_ALLOC(medium) ((SDPL_LAYER(medium)->SDPL_map) != NULL)
#define FREE_DEV(medium) do { } while (0)
#define INVALID_REQUEST(p, dev) (FALSE)

#else /* CONFIG_BLK_SDPL_EMBEDDED */

#define SDPL_LAYER(medium) (&SDPL_layers[medium])
#define INIT_DEVS(medium) \
do { SDPL_layers[medium].SDPL_dev = NULL; } while (0)
#define FOR_EACH_MEDIUM  for (medium = 0; medium < (MAX_MEDIUM); medium++)
#define IS_DEV_ALLOC(medium) (SDPL_layers[medium].SDPL_dev != NULL)
#define FREE_DEV(medium)                 \
do {                                     \
    kfree(SDPL_layers[medium].SDPL_dev); \
    SDPL_layers[medium].SDPL_dev = NULL;
} while (0)
#define INVALID_REQUEST(p, dev) ((p) == NULL || MAJOR(dev) != SDPL_MAJOR)

#endif /* CONFIG_BLK_SDPL_EMBEDDED */

#define SDPL_DEV(medium, minor) ((SDPL_LAYER(medium))->SDPL_dev[minor])

#if CONFIG_BLK_SDPL_STATS

#define MOD_INIT_OPENS(medium, minor)                          \
do { SDPL_DEV(medium, minor).opens = 0; } while (0)
#define MOD_DEC_OPENS(medium, minor)                           \
do {                                                           \
     SDPL_DEV(medium, minor).opens--;                          \
     sdpl_debug("SDPL opens: %ld for device %d-%d\n",          \
                SDPL_DEV(medium, minor).opens, medium, minor); \
} while(0)
#define MOD_INC_OPENS(medium, minor)                           \
do {                                                           \
     SDPL_DEV(medium, minor).opens++;                          \
     sdpl_debug("SDPL opens: %ld for device %d-%d\n",          \
                SDPL_DEV(medium, minor).opens, medium, minor); \
} while(0)
#define MOD_INIT_READS(medium, minor)                          \
do { SDPL_DEV(medium, minor).reads = 0; } while (0)
#define MOD_INC_READS(medium, minor)                  \
do { SDPL_DEV(medium, minor).reads++; } while (0)
#define MOD_INIT_WRITES(medium, minor)                \
do { SDPL_DEV(medium, minor).writes = 0; } while (0)
#define MOD_INC_WRITES(medium, minor)                 \
do { SDPL_DEV(medium, minor).writes++; } while (0)

#else /* CONFIG_BLK_SDPL_STATS */

#define MOD_INIT_OPENS(medium, minor) do { } while (0)
#define MOD_DEC_OPENS(medium, minor) do { } while (0)
#define MOD_INC_OPENS(medium, minor) do { } while (0)
#define MOD_INIT_READS(medium, minor) do { } while (0)
#define MOD_INC_READS(medium, minor) do { } while (0)
#define MOD_INIT_WRITES(medium, minor) do { } while (0)
#define MOD_INC_WRITES(medium, minor) do { } while (0)

#endif /* CONFIG_BLK_SDPL_STATS */

#if NO_PERM_CHECK
#define CHECK_PERMS do {} while(0)
#else /* !NO_PERM_CHECK */
#define CHECK_PERMS \
if (!capable(CAP_SYS_ADMIN)) { ret = -EACCES; break; }
#endif /* !NO_PERM_CHECK */

/*
 * Debug level:
 * 0 -> No messages
 * 1 -> Error messages
 * 2 -> Info  messages
 * 3 -> Debug messages
 * Note: Debug mesages are always KERN_CRIT,
 *       so we're sure to always get them...
 */
/* Messages handler definitions */
#if CONFIG_BLK_SDPL_DEBUG > 2
#define sdpl_debug(format, args...) \
do { printk(KERN_CRIT format, ##args); } while (0)
#else /* CONFIG_BLK_SDPL_DEBUG > 2 */
#define sdpl_debug(format, args...) do { } while (0)
#endif /* CONFIG_BLK_SDPL_DEBUG > 2 */
#if CONFIG_BLK_SDPL_DEBUG > 0
#define sdpl_error(format, args...) \
do { printk(KERN_ERR format, ##args); } while (0)
#else /* CONFIG_BLK_SDPL_DEBUG > 0 */
#define sdpl_error(format, args...) do { } while (0)
#endif /* CONFIG_BLK_SDPL_DEBUG > 0 */
#if CONFIG_BLK_SDPL_DEBUG > 1
#define sdpl_info(format, args...) \
do { printk(KERN_INFO format, ##args); } while (0)
#else /* CONFIG_BLK_SDPL_DEBUG > 1 */
#define sdpl_info(format, args...) do { } while (0)
#endif /* CONFIG_BLK_SDPL_DEBUG > 1 */

#ifdef CONFIG_PROC_FS
int sdpl_get_layers(char *page, char **start, off_t off,
                    int count, int *eof, void *data);
#endif

#if (LINUX_VERSION_CODE < 0x2040A)

typedef struct { kdev_t device; } block_dev_t;
#define ALLOC_BLOCKDEV(bdev, indev)                   \
do {                                                  \
     bdev = kmalloc(sizeof(block_dev_t), GFP_KERNEL); \
     if (bdev != NULL) bdev->device = (indev);        \
} while (0)
#define RELEASE_BLOCKDEV(bdev) do { kfree(bdev); } while (0)

static inline void get_bh (struct buffer_head * bh)
{
    atomic_inc(&(bh)->b_count);
}

#else /* (LINUX_VERSION_CODE < 0x2040A) */

typedef struct block_device block_dev_t;
#define ALLOC_BLOCKDEV(bdev, indev)                   \
do {                                                  \
    (bdev) = bdget(kdev_t_to_nr(indev));              \
} while(0)
#define RELEASE_BLOCKDEV(bdev) do { bdput(bdev); } while (0)

#endif /* (LINUX_VERSION_CODE < 0x2040A) */

static inline void put_dev_sector(Sector p)
{
    page_cache_release(p.v);
}

void *sdpl_realloc (void *ptr, int old_size, int new_size);

#define BUFFER(sector, nsect) \
              ((unsigned char *)page_address(sector->v) + \
               512 * (nsect % (PAGE_CACHE_SIZE / 512)))

#else /* __KERNEL__ */

/* This is a kludge to use the same read function in kernel and user space */
typedef struct { unsigned char *sector; } Sector;

/* User space functions definitions */
#define sdpl_malloc(a) ({ malloc(a); })
#define sdpl_free(a) ({ free(a); })
#define sdpl_realloc(ptr, old_size, new_size) ({ realloc(ptr, new_size); })
#if CONFIG_BLK_SDPL_DEBUG > 1
#define sdpl_debug(format, args...) \
do { fprintf(stderr, format, ##args); } while (0)
#else /* CONFIG_BLK_SDPL_DEBUG > 1 */
#define sdpl_debug(format, args...) do { } while (0)
#endif /* CONFIG_BLK_SDPL_DEBUG > 1 */
#define sdpl_error(format, args...) \
do { fprintf(stderr, format, ##args); } while (0)
#define sdpl_info(format, args...)  \
do { fprintf(stdout, format, ##args); } while (0)

/*
 * These functions should be redirect to 64 bits one by glibc,
 * if needed
 */
#define sdpl_open open
#define sdpl_lseek lseek
#define sdpl_read read
#define sdpl_write write
#if CONFIG_BLK_SDPL_64BITS
#define sdpl_stat(file, stat) ({ lstat64(file, stat); })
typedef struct stat64 sdpl_stat_t;
#else /* CONFIG_BLK_SDPL_64BITS */
#define sdpl_stat(file, stat) ({ lstat(file, stat); })
typedef struct stat sdpl_stat_t;
#endif /* CONFIG_BLK_SDPL_64BITS */

typedef struct block_device { int fd; } block_dev_t;
typedef block_dev_t *kdev_t;

static inline void put_dev_sector(UNUSED Sector p) {}

#define BUFFER(sect, nsect) (sect->sector /* + nsect * 512 */)
#define ALLOC_BLOCKDEV(bdev, indev) do { (bdev) = (indev); } while (0)
#define RELEASE_BLOCKDEV(bdev) do {} while (0)

#if !CONFIG_BLK_SDPL_EMBEDDED
void fix_bflags (struct SDPL_map *SDPL_map);
void fix_SDPL_infos (struct SDPL_map *SDPL_map);
unsigned long get_ressize (unsigned long tot_size, int size, int info_size,
                           int embedded, std_type *pe_size,
                           std_type *pm_size, std_type *bm_size);
int get_best_size (const unsigned long tot_size, unsigned long *reserved,
                   const int min_blocs, const int max_blocs, int *nblocs,
                   int info_size, int embedded);
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */

static inline int write_dev_sector (kdev_t bdev, unsigned long n, Sector *sect)
{
    sdpl_lseek(bdev->fd, n * 512, SEEK_SET);
    return sdpl_write(bdev->fd, sect->sector, 512);
}

static inline void fsync_dev (kdev_t bdev)
{
    fsync(bdev->fd);
}

/* From edit.c */
void interract (struct SDPL_map *pSDPL_map, kdev_t bloc_dev,
                Sector sect);

#endif /* __KERNEL__ */

/* Definitions shared between sdpledit / sdpldump */
#if CONFIG_BLK_SDPL_EDIT
#define EDIT_CMD
#define EDIT_VAR
#else /* CONFIG_BLK_SDPL_EDIT */
#define EDIT_CMD static
#define EDIT_VAR static
#endif /* CONFIG_BLK_SDPL_EDIT */

/* Usefull routines for 64 bits datas manipulation */
#if CONFIG_BLK_SDPL_64BITS
u64 ntohll (u64 data);
u64 htonll (u64 data);
#endif /* CONFIG_BLK_SDPL_64BITS */

/* Get the same API for kernel and user space functions */
unsigned char *read_dev_sector (block_dev_t *bdev,
                                unsigned long n, Sector *p);
void get_parts_infos (struct SDPL_map *SDPL_map, Sector sect,
                      block_dev_t *bdev);

std_type count_free_blocs (struct SDPL_map *SDPL_map);
int get_SDPL_infos (struct SDPL_map *SDPL_map);

#if CONFIG_BLK_SDPL_DEBUG > 2 || !defined(__KERNEL__)
void dump_SDPL_infos (struct SDPL_map *SDPL_map);
void dump_part_infos (struct SDPL_map *SDPL_map, int minor);
void dump_free_blocks (struct SDPL_map *SDPL_map);
#endif /* CONFIG_BLK_SDPL_DEBUG > 2 || !defined(__KERNEL__) */

/* From libedit.c */
int add_spartition (struct SDPL_map *SDPL_map, std_type nb_blocs);
int remove_spartition (struct SDPL_map *SDPL_map, std_type part_num);
int remove_bloc_from_part (struct SDPL_map *SDPL_map, std_type part_num,
                           std_type num_bloc);
int add_blocs_to_part (struct SDPL_map *SDPL_map, std_type part_num,
                       std_type nb_blocs);
void mark_partition_bootable (struct SDPL_map *SDPL_map, std_type part_num);
void toggle_flag (u16 *flag, int toggle);
void toggle_part_flag (struct SDPL_map *SDPL_map, std_type part_num,
                       int toggle);
static inline void toggle_bloc_flag (struct SDPL_map *SDPL_map,
                                     std_type bloc_num,
                                     int toggle)
{
    toggle_flag(&SDPL_map->bflags[bloc_num], toggle);
}
void save_sdpl (struct SDPL_map *SDPL_map, kdev_t bdev,
                Sector *sector, std_type *cur_sect);

#endif /* __SDPL_H__ */
