/*
 * SDPL sdpldump.c: a dump utility for sdpl
 *
 *   Copyright (C) 2001, 2002, 2003, 2004, Jocelyn MAYER <l_indien@magic.fr>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Simple Disk Partition Layer
 *
 * User mode dump program
 *
 * This is a simple programm which dump the layer
 * from a file or a block device
 *
 * Usage: sdpldump [-k] file
 *
 */

/**SECTION: sysutils */
/**SECTION: sdpl */
/**SECTION: edit tool */

#include <stdlib.h>
#include <stdio.h>
#ifdef CONFIG_BLK_SDPL_64BITS
/* These are the GLIBC definition to use 64 bits file support */
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 1
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 1
#endif
#else /* CONFIG_BLK_SDPL_64BITS */
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 0
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 0
#endif
#endif /* CONFIG_BLK_SDPL_64BITS */
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mount.h>

#include <linux/types.h>
#include <linux/version.h>
#include <linux/sdpl.h>

/** Display usage
 */
static void usage (char *name)
{
    fprintf(stderr, "Usage: %s file\n", name);
}

/** Main entry point
 */
int main (int argc, char **argv)
{
    struct block_device bdev;
    sdpl_stat_t f_stat;
    Sector sect;
    std_type i;
    /* main structure */
    struct SDPL_map *SDPL_map = NULL;
    /* File description */
    char *filename;
    /* Flags */
#if !CONFIG_BLK_SDPL_EMBEDDED
    int ret = -1;
#endif

    if (argc <  2) {
        usage(argv[0]);
        exit(1);
    }

    bdev.fd = -1;
    filename = argv[1];
    if (argc !=  2) {
        usage(argv[0]);
        exit(1);
    }
    if ((sect.sector = malloc(512)) == NULL)
        goto error;
    if ((SDPL_map = malloc(sizeof(struct SDPL_map))) == NULL)
        goto error;
#if CONFIG_BLK_SDPL_EMBEDDED
    if ((bdev.fd = sdpl_open(filename, O_RDONLY | O_EXCL | O_SYNC)) < 0 ||
        sdpl_stat(filename, &f_stat) < 0) {
        goto error;
    }
#else /* CONFIG_BLK_SDPL_EMBEDDED */
    if ((bdev.fd = 
         sdpl_open(filename, O_RDWR | O_EXCL | O_SYNC | O_LARGEFILE)) < 0 ||
        sdpl_stat(filename, &f_stat) < 0) {
        goto error;
    }
    SDPL_map->infos = NULL;
    SDPL_map->pentries = NULL;
    SDPL_map->pmagic = NULL;
    SDPL_map->pflags = NULL;
    SDPL_map->psize = NULL;
    SDPL_map->poffset = NULL;
    SDPL_map->pnext = NULL;
    SDPL_map->pbloc = NULL;
    SDPL_map->bflags = NULL;
#endif /* CONFIG_BLK_SDPL_EMBEDDED */

    /* Get the MBR */
    if ((SDPL_map->mbr = 
         (SDPL_br_t *)read_dev_sector(&bdev, 0, &sect)) == NULL) {
        sdpl_error("Can't read boot sector\n");
        goto error;
    }
    SDPL_map->res_size = 1024;
    SDPL_map->device_size = f_stat.st_size;
    if (f_stat.st_size == 0 && S_ISBLK(f_stat.st_mode)) {
        int size;
        if (ioctl(bdev.fd, BLKGETSIZE, &size) < 0 || size < 0) {
            fprintf(stderr, "Unbale to get partition size\n");
            goto end;
        }
        SDPL_map->device_size = size;
    }
    if (SDPL_map->device_size < 2048) {
        sdpl_error("Device too small: " STD_ATTRIB " bytes (2048)\n",
                   SDPL_map->device_size);
        goto error;
    }
    SDPL_map->st_size = f_stat.st_size - SDPL_map->res_size;
    if (get_SDPL_infos(SDPL_map) < 0) { 
#if CONFIG_BLK_SDPL_EMBEDDED
        goto error;
#else /* CONFIG_BLK_SDPL_EMBEDDED */
        //        fix_SDPL_infos(SDPL_map);
        goto go_on;
#endif /* CONFIG_BLK_SDPL_EMBEDDED */
    }
    dump_SDPL_infos(SDPL_map);

    get_parts_infos(SDPL_map, sect, &bdev);
    for (i = 0; i < NPARTS(SDPL_map); i++) {
        dump_part_infos(SDPL_map, i);
    }
    dump_free_blocks(SDPL_map);
#if !CONFIG_BLK_SDPL_EMBEDDED
 go_on:
    interract(SDPL_map, &bdev, sect);
    ret = 0;
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
    goto end;

 error:
    if (sect.sector != NULL)
        sdpl_free(sect.sector);
    perror(argv[0]);

 end:

    if (bdev.fd > 0)
        close(bdev.fd);

    return -1;
}
